/* Program name --- const.c
 *       written by Sakurai   2003.7.30
 */

#include        <stdio.h>
#include        <stdlib.h>
#include        <string.h>
#include        <math.h>
#include        <time.h>
#include        "approxdp.h"
#include        "lbs.h"

#define         DBG         0

void error(char *);
extern int   seq_read();

/* for memory allocation */
extern void malloc_float();
extern void malloc_int();
extern float *allot_float();
extern int *allot_int();
float *buf_float;
int   *buf_int;
int   float_cnt, int_cnt;
int   limit_float, limit_int;
/* for memory allocation */

int main(argc,argv)
int	argc;
char	**argv;
{
  struct  SeqData **sqd, *buf_sqd;
  struct  LbsData **lbsd, *buf_lbsd;
  int        sqd_cnt=0, lbsd_cnt=0;
  int        dblength, dbsize, upper_len;
  int        i,j,k,p,n;
  int        segs1[CALCS], calcs, c1=C1, total_segs1;
  int        diff_flg=OFF;
  char       dbfile[128], filename[128];
  char       read_buf[BUFSIZE];
  FILE       *fp;
  float      dist;

  if(argc != 4){
    error(argv[0]);
    exit(1);
  }
  system("/bin/date");

  strcpy(dbfile,argv[1]);
  dblength = atoi(argv[2]);
  strcpy(filename,argv[3]);

  calcs = setcalcs(dblength);
  setsegs(segs1, dblength, calcs);
  total_segs1=0;
  printf("Length of the approximate sequences:\n");
  for (k=0 ; k<calcs ; k++){
    printf("(level %d) %d\n", k, segs1[k]);
    total_segs1 += segs1[k];
  }

  /*
   * Read data
   */
  if ( ( fp = fopen ( dbfile , "r" ) ) == NULL ){
    printf( "  %s can not open\n" , dbfile);
    exit(1);
  }
  j = 0;
  while (fgets(read_buf,BUFSIZE,fp) != NULL )
    j++;
  dbsize = j;
  rewind(fp);

  sqd = (struct SeqData **)malloc(sizeof(struct SeqData *) * dbsize);
  buf_sqd = (struct SeqData *)malloc(sizeof(struct SeqData) * dbsize);
  if (sqd == NULL || buf_sqd == NULL){
    printf("Allocation error : sqd, buf_sqd\n");
    exit(1);
  }
  lbsd = (struct LbsData **)malloc(sizeof(struct LbsData *) * dbsize);
  buf_lbsd = (struct LbsData *)malloc(sizeof(struct LbsData) * dbsize);
  if (lbsd == NULL || buf_lbsd == NULL){
    printf("Allocation error : lbsd\n");
    exit(1);
  }

  j = 0;
  /* sqd */
  j += dbsize * dblength;
  /* lbsd */
  j += dbsize * total_segs1 * 2;

  malloc_float(j);

  for (i=0 ; i<dbsize ; i++) {
    sqd[i] = buf_sqd + sqd_cnt;
    sqd_cnt++;
    sqd[i]->code = allot_float(dblength);
  }
  for (i=0 ; i<dbsize ; i++) {
    lbsd[i] = buf_lbsd + lbsd_cnt;
    lbsd_cnt++;
    for (k=0 ; k<calcs ; k++) {
      lbsd[i]->segd[k].lower = allot_float(segs1[k]);
      lbsd[i]->segd[k].upper = allot_float(segs1[k]);
    }
  }

  p = 0;
  diff_flg = OFF;
  while (fgets(read_buf,BUFSIZE,fp) != NULL ) {
    sqd[p]->len = seq_read( read_buf, sqd[p]->code, dblength);
    if (dblength != sqd[p]->len)
      diff_flg = ON;
    p++;
  }
  fclose(fp);

  /*
   * Compute the segment of each sequence
   */
  for (i=0 ; i<dbsize ; i++){
    lbsd[i]->seq_length = sqd[i]->len;
    setsegs(segs1, lbsd[i]->seq_length, calcs);
    for (k=0 ; k<calcs ; k++){
      lbsd[i]->segd[k].segs = segs1[k];
      getpaa(sqd[i]->code, lbsd[i], k);
    }
  }

  fp = fopen(filename,"wb");
  if (fp == NULL){
    fprintf(stderr,"file open error! %s\n", filename);
    exit(1);
  }

  /*
   * Create the index file
   */
  n=fwrite(&(dbsize), sizeof(int), 1, fp);
  n=fwrite(&(dblength), sizeof(int), 1, fp);
  n=fwrite(&(diff_flg), sizeof(int), 1, fp);
  n=fwrite(&(calcs), sizeof(int), 1, fp);

  for ( i=0 ; i < dbsize ; i++){
    n=fwrite(&(lbsd[i]->seq_length), sizeof(int), 1, fp);
    for (k=0 ; k<calcs ; k++){
      n=fwrite(lbsd[i]->segd[k].lower,  sizeof(float), 
	       lbsd[i]->segd[k].segs, fp);
      n=fwrite(lbsd[i]->segd[k].upper, sizeof(float), 
	       lbsd[i]->segd[k].segs, fp);
    }
  }
  fclose(fp);
  free(buf_float);
  free(buf_int);
  system("/bin/date");
}


void error(char *toolname)
{
    fprintf(stderr,"Usage : %s  db_file  max_length  index_file\n",toolname);
    exit(1);
}
