/******************************************************************************
* File name     : The definition of Dymamic Programming Module
* Date          : December 6, 2005
* Note          : Definition of Macros, tables, function-like macros, functions
******************************************************************************/

#include <limits.h>

  /*--------------------------------------------------------------------------*
   * Marco definition
   *--------------------------------------------------------------------------*
   */

#define INITIALIZE_VALUE 1e20   /* Initial distance value                    */
#define DIAGONAL_WEIGHT     1
#define YES                 1
#define NO                  0

  /*
   * Warping scope
   */
#define ITAKURA            NO
// #define ITAKURA           YES
#define SC_WIDTH            1
// #define SC_WIDTH          0.1
#define ITAKURA_WIDTH       2

  /*--------------------------------------------------------------------------*
   * Table definition
   *--------------------------------------------------------------------------*
   */

  /*
   *   Management table
   *   keeps the intermediate results of distance computation for dynamic 
   *   programming
   */

typedef struct _dp_match {
  size_t  buffer_size;
  struct _dp_match_opt_value *result1;
  struct _dp_match_opt_value *result2;
  int    seq1_length;
  int    seq2_length;
  int    *start_point;
  int    *end_point;
} dp_match_t;

  /*
   * Optimum distance value during processing of DTW
   */
struct _dp_match_opt_value {
  float optimal_value;
};


  /*--------------------------------------------------------------------------*
   * Definition of function-like macros
   *--------------------------------------------------------------------------*
   */

  /*
   * For computing the distance at the current grid point
   */
#define CALC_DISTANCE(SP1, SP2) \
  ((SP1 - SP2) * (SP1 - SP2))

  /*
   * For setting the sequence lengths
   */
#define dpMatch_set_sequence_length(self, seq1, seq2) \
  (self)->seq1_length = seq1; \
  (self)->seq2_length = seq2

  /*--------------------------------------------------------------------------*
   * Functions
   *--------------------------------------------------------------------------*
   */

dp_match_t* dpMatch_malloc(
  size_t      buffer_size1,
  size_t      buffer_size2
);
void dpMatch_initialization(
  dp_match_t* self,
  int len1,
  int len2
);
void dpMatch_clear(
  dp_match_t* self
);
float dpMatch_distance(
  dp_match_t* self,
  float      *sequence1,
  float      *sequence2
);
float dpMatch_threshold(
  dp_match_t* self,
  float      *sequence1,
  float      *sequence2,
  float      threshold
);
float dpMatch_constraint(
  dp_match_t* self,
  float      *sequence1,
  float      *sequence2
);
