/* Program name --- exactdp.c
 *       written by Sakurai   2002.10.11
 */

#include        <stdio.h>
#include        <stdlib.h>
#include        <string.h>
#include        <math.h>
#include        <time.h>
#include        "approxdp.h"
#include        "dpmatch.h"

#define         DBG         0
#define         DBG2        0
#define         DBG3        0
#define         RESULT      1

extern int   seq_read();

void error(char *);
void malloc_read();
int change();

/* for memory allocation */
extern void malloc_float();
extern void malloc_int();
extern float *allot_float();
extern int *allot_int();
float *buf_float;
int   *buf_int;
int   float_cnt, int_cnt;
int   limit_float, limit_int;
/* for memory allocation */

struct	SortData {
  int      num;
  struct  NetData *netd;
};
struct	NetData {
  int      id;
  float   distance;
};

int     dbsize, qsize;
struct  SeqData **sqd, **query, *buf_sqd;
struct  SortData buf_sd[1];
struct  SortData *nn=buf_sd;
struct  NetData  *buf_net;
int     *que_len, *seq_len;

main(argc,argv)
int	argc;
char	*argv[];
{

  int     i, j, m;
  int     num_nn, dblength, qlength;
  char    dbfile[128], queryfile[128];
  float   dist;

  dp_match_t *dpm;

  clock_t    timecnt;
  double     totaltime=0.0;

  if(argc != 6){
    error(argv[0]);
    exit(1);
  }

  strcpy(dbfile,argv[1]);
  strcpy(queryfile,argv[2]);
  dblength = atoi(argv[3]);
  qlength = atoi(argv[4]);
  num_nn = atoi(argv[5]);

  malloc_read(dbfile, queryfile, dblength, qlength);
  dpm = dpMatch_malloc(dblength, qlength);

  /*
   * exact DTW
   */
  system("/bin/date");
  for (j=0 ; j<qsize ; j++){
    timecnt = clock();
    m = num_nn - 1;
    for (i=0 ; i<dbsize ; i++){
      dpMatch_initialization(dpm, seq_len[i], que_len[j]);
      dist = dpMatch_constraint(dpm, sqd[i]->code, query[j]->code);
      if (i >= num_nn){
	if (dist < nn->netd[m].distance){
	  nn->netd[m].distance = dist;
	  nn->netd[m].id = i;
	  qsort(nn->netd, num_nn, sizeof(struct NetData), change);
	}
      }
      else{
	nn->netd[i].distance = dist;
	nn->netd[i].id = i;
	qsort(nn->netd, i+1, sizeof(struct NetData), change);
      }
    }
    totaltime += (double)(clock() - timecnt) / (double)CLOCKS_PER_SEC;
#if RESULT
    printf("Sequence ID: %d\n", j);
    for (i=0 ; i<num_nn ; i++) 
      printf("(%d) id=%d, distance=%f\n", i+1, nn->netd[i].id, 
	     nn->netd[i].distance);
#endif
  }
  system("/bin/date");
  totaltime /= (float)qsize;
  printf("%.4f sec.\n",totaltime);
  printf("\n");
  dpMatch_clear(dpm);
}

void malloc_read(dbfile, queryfile, dblength, qlength)
char *dbfile;
char *queryfile;
int  dblength;
int  qlength;
{

  int     i, p;
  int     sqd_cnt=0;
  char    read_buf[BUFSIZE];
  FILE    *fp1, *fp2;

  /*
   * Count the number of sequences
   */
  if ( ( fp1 = fopen ( dbfile , "r" ) ) == NULL ){
    printf( "  %s can not open\n" , dbfile);
    exit(1);
  }
  p = 0;
  while (fgets(read_buf,BUFSIZE,fp1) != NULL )
    p++;
  dbsize = p;
  rewind(fp1);
  if ( ( fp2 = fopen ( queryfile , "r" ) ) == NULL ){
    printf( "  %s can not open\n" , queryfile);
    exit(1);
  }
  p = 0;
  while (fgets(read_buf,BUFSIZE,fp2) != NULL )
    p++;
  qsize = p;
  rewind(fp2);

  /*
   * memory allocation (sqd)
   */
  sqd = (struct SeqData **)malloc(sizeof(struct SeqData *) * dbsize);
  query = (struct SeqData **)malloc(sizeof(struct SeqData *) * qsize);
  buf_sqd = (struct SeqData *)malloc(sizeof(struct SeqData)
				     * (dbsize+qsize));
  if (sqd == NULL || query == NULL || buf_sqd == NULL){
    printf("Allocation error : sqd, query, buf_sqd\n");
    exit(1);
  }

  /*
   * memory allocation (int)
   */
  malloc_int(dbsize + qsize);
  seq_len = allot_int(dbsize);
  que_len = allot_int(qsize);

  /*
   * memory allocation (float)
   */
  malloc_float(dbsize * dblength + qsize * qlength);
  for (i=0 ; i<dbsize ; i++) {
    sqd[i] = buf_sqd + sqd_cnt;
    sqd_cnt++;
    sqd[i]->code = allot_float(dblength);
  }
  for (i=0 ; i<qsize ; i++) {
    query[i] = buf_sqd + sqd_cnt;
    sqd_cnt++;
    query[i]->code = allot_float(qlength);
  }

  /*
   * memory allocation (nn)
   */
  buf_net = (struct NetData *)malloc(sizeof(struct NetData) * dbsize);
  if (buf_net == NULL){
    printf("Allocation error : buf_net\n");
    exit(1);
  }
  nn->netd = buf_net;

  /*
   * Read the data and query sequence files
   */
  p=0;
  while (fgets(read_buf,BUFSIZE,fp1) != NULL ) {
    i = seq_read( read_buf, sqd[p]->code, dblength);
    seq_len[p] = i;
    p++;
  }
  fclose(fp1);
  p=0;
  while (fgets(read_buf,BUFSIZE,fp2) != NULL ) {
    i = seq_read( read_buf, query[p]->code, qlength);
    que_len[p] = i;
    p++;
  }
  fclose(fp2);
}

int change(nd1,nd2)
struct NetData *nd1;
struct NetData *nd2;
{
  if (nd1->distance > nd2->distance) return  1;
  if (nd1->distance < nd2->distance) return -1;
  return  0;
}

void error(char *toolname)
{
  fprintf(stderr, 
    "Usage : %s  dbfile  queryfile  dblength  query_length  num_neighbors\n", 
    toolname);
  exit(1);
}
