/* Program name --- search.c
 *       written by Sakurai   2003.8.15
 */

#include        <stdio.h>
#include        <stdlib.h>
#include        <string.h>
#include        <math.h>
#include        <time.h>
#include        "approxdp.h"
#include        "lbs.h"
#include        "dpmatch.h"

#define         DBG           0
#define         ACCESS        0
#define         RESULT        1

#define         PAGESIZE    8192
#define         FLOAT_SIZE     4

extern int   seq_read();

/* for memory allocation */
extern void malloc_float();
extern void malloc_int();
extern float *allot_float();
extern int *allot_int();
float *buf_float;
int   *buf_int;
int   float_cnt, int_cnt;
int   limit_float, limit_int;
/* for memory allocation */

void malloc_read();
void nnsearch_m();
void accinfo();
int  change();
void error(char *) ;

struct  SeqData **sqd, **qsqd, *buf_sqd;
struct  WarpData buf_warpd[1], *warpd=buf_warpd;
struct  LbsData **lbsd, *buf_lbsd;
struct  LbsData buf_qlbsd[1], *qlbsd=buf_qlbsd;
int     calcs, segs1[CALCS], segs2[CALCS], c1=C1;
int     dblength, qlength, dbsize, diff_flg=OFF;
int     *apid;
float   *work2, *apdist;

dp_match_t *dpm;

struct	SortData {
                int      num;
                struct  NetData *netd;
	};

struct	NetData {
                int      id;
                float    distance;
	};

struct  SortData buf_sd[2];
struct  SortData *nn=buf_sd;
struct  SortData *vnn=buf_sd+1;
struct  NetData  *buf_net;

int     num_nn;
int     qsize;
int     *prt_buf1;
float   *prt_buf2;

int     n_oid, t_oid=0, t_pid[CALCS];

int main(argc,argv)
int	argc;
char	*argv[];
{

  int        i,j,k,m,p;
  char       dbfile[128];
  char       index_file[128];
  char       query_file[128];
  int        query_cnt;

  clock_t timecnt;
  double  stime=0, totaltime=0.0;

#if ACCESS
  char       acc_file[128];
#endif

#if ACCESS
  if (argc != 7 ) error(argv[0]);
  strcpy(acc_file,argv[6]);
#else
  if (argc != 6 ) error(argv[0]);
#endif
  strcpy(index_file,argv[1]);
  strcpy(dbfile,argv[2]);
  strcpy(query_file,argv[3]);
  qlength = atoi(argv[4]);
  num_nn = atoi(argv[5]);

  /*
   * Memory allocation and data loading
   */
  malloc_read(dbfile, index_file, query_file);

  /*
   * k-nearest neighbor processing
   */

#if ACCESS
  t_pid[0] = dbsize * qsize;
  for (k=1 ; k<calcs ; k++)
    t_pid[k] = 0;
#endif

  system("/bin/date");

  for ( query_cnt=0 ; query_cnt<qsize ; query_cnt++ ) {

    timecnt = clock();

#if ACCESS
    n_oid = 0;
#endif

    nnsearch_m(query_cnt);

    stime += (double)(clock() - timecnt);

    /*
     * copy the search result
     */
    for (i=0 ; i<num_nn ; i++) {
      prt_buf1[query_cnt*num_nn+i] = nn->netd[i].id;
      prt_buf2[query_cnt*num_nn+i] = nn->netd[i].distance;
    }
    
#if ACCESS
    t_oid += n_oid;
#endif
  }

  totaltime = stime / (double)CLOCKS_PER_SEC / (double)qsize;
  printf("%.4f sec.\n",totaltime);
  system("/bin/date");

#if RESULT
  /*
   * report all the search results
   */
  for (i=0 ; i<qsize ; i++) {
    printf("Sequence ID: %d\n", i);
    for (j=0 ; j<num_nn ; j++) {
      printf("(%d) id=%d, distance=%f\n", j+1, prt_buf1[i*num_nn+j], 
	     prt_buf2[i*num_nn+j]);
    }
  }
#endif

#if ACCESS
  /*
   * report the number of page accesses
   */
  accinfo(acc_file);
#endif

  dpMatch_clear(dpm);

  free(buf_int);
  free(buf_float);
  free(buf_lbsd);
  free(buf_sqd);
  free(buf_net);
}

void malloc_read(dbfile, index_file, query_file)
char *dbfile;
char *index_file;
char *query_file;
{

  int        n, k, i, j, p;
  int        sqd_cnt=0, lbsd_cnt=0;
  char       read_buf[BUFSIZE];
  int        total_segs1, total_segs2;
  FILE       *fp, *fp2;

  /*
   * read the header of the index file
   */
  if ( ( fp = fopen ( index_file , "rb" ) ) == NULL ){
    printf( "  %s can not open\n" , index_file);
    exit(1);
  }
  n=fread(&(dbsize), sizeof(int), 1, fp);
  n=fread(&(dblength), sizeof(int), 1, fp);
  n=fread(&(diff_flg), sizeof(int), 1, fp);
  n=fread(&(calcs), sizeof(int), 1, fp);
  setsegs(segs1, dblength, calcs);
  for (k=0 ; k<calcs ; k++)
    total_segs1 += segs1[k];
  setsegs(segs2, qlength, calcs);
  total_segs2 = 0;
  for (k=0 ; k<calcs ; k++)
    total_segs2 += segs2[k];

  /*
   * count the number of queries
   */
  if ( ( fp2 = fopen ( query_file , "r" ) ) == NULL ){
    printf( "  %s can not open\n" , query_file);
    exit(1);
  }
  qsize = 0;
  while (fgets(read_buf,BUFSIZE,fp2) != NULL )
    qsize++;
  rewind(fp2);

  /*
   * memory allocation (buf_int)
   */
  j = 0;
  /* prt_buf1 */
  j += qsize*num_nn;
  /* apid  */
  j += dbsize;
  /* warpd */
  j += total_segs1 * 3 + total_segs2;

  malloc_int(j);

  apid = allot_int(dbsize);
  prt_buf1 = allot_int(qsize*num_nn);

  for (k=0 ; k<calcs ; k++){
    warpd->seg_env[k].begin = allot_int(segs1[k]);
    warpd->seg_env[k].end = allot_int(segs1[k]);
    warpd->ed[k].elms1 = allot_int(segs1[k]);
    warpd->ed[k].elms2 = allot_int(segs2[k]);
  }

  /*
   * memory allocation (buf_float)
   */
  j = 0;
  /* sqd and qsqd */
  j = dbsize * dblength + qsize * qlength;
  /* prt_buf2 */
  j += qsize*num_nn;
  /* lbsd */
  j += dbsize * total_segs1 * 2;
  /* qlbsd */
  j += total_segs2 * 2;
  /* work2 */
  j += qlength*2;
  /* apdist */
  j += dbsize;

  malloc_float(j);
  prt_buf2 = allot_float(qsize*num_nn);
  work2 = allot_float(qlength*2);
  apdist = allot_float(dbsize);

  /*
   * memory allocation (lbsd)
   */
  lbsd = (struct LbsData **)malloc(sizeof(struct LbsData *) * dbsize);
  buf_lbsd = (struct LbsData *)malloc(sizeof(struct LbsData) * dbsize);
  if (lbsd == NULL || buf_lbsd == NULL){
    printf("Allocation error : lbsd\n");
    exit(1);
  }
  for (i=0 ; i<dbsize ; i++) {
    lbsd[i] = buf_lbsd + lbsd_cnt;
    lbsd_cnt++;
    for (k=0 ; k<calcs ; k++) {
      lbsd[i]->segd[k].lower = allot_float(segs1[k]);
      lbsd[i]->segd[k].upper = allot_float(segs1[k]);
    }
  }
  for (k=0 ; k<calcs ; k++) {
    qlbsd->segd[k].lower = allot_float(segs2[k]);
    qlbsd->segd[k].upper = allot_float(segs2[k]);
  }

  /*
   * memory allocation (sqd)
   */
  sqd = (struct SeqData **)malloc(sizeof(struct SeqData *) * dbsize);
  qsqd = (struct SeqData **)malloc(sizeof(struct SeqData *) * qsize);
  buf_sqd = (struct SeqData *)malloc(sizeof(struct SeqData) * (dbsize+qsize));
  if (sqd == NULL || qsqd == NULL || buf_sqd == NULL){
    printf("Allocation error : sqd, qsqd, buf_sqd\n");
    exit(1);
  }
  for (i=0 ; i<dbsize ; i++) {
    sqd[i] = buf_sqd + sqd_cnt;
    sqd_cnt++;
    sqd[i]->code = allot_float(dblength);
  }
  for (i=0 ; i<qsize ; i++) {
    qsqd[i] = buf_sqd + sqd_cnt;
    sqd_cnt++;
    qsqd[i]->code = allot_float(qlength);
  }

  /*
   * read the index file
   */
  for ( i=0 ; i < dbsize ; i++){
    n=fread(&(lbsd[i]->seq_length), sizeof(int), 1, fp);
    setsegs(segs1, lbsd[i]->seq_length, calcs);
    for (k=0 ; k<calcs ; k++){
      lbsd[i]->segd[k].segs = segs1[k];
      n=fread(lbsd[i]->segd[k].lower,  sizeof(float), 
	      lbsd[i]->segd[k].segs, fp);
      n=fread(lbsd[i]->segd[k].upper, sizeof(float), 
	      lbsd[i]->segd[k].segs, fp);
    }
  }
  fclose(fp);

  /*
   * read the query file
   */
  p=0;
  while (fgets(read_buf,BUFSIZE,fp2) != NULL ) {
    qsqd[p]->len = seq_read( read_buf, qsqd[p]->code, qlength);
    p++;
  }
  fclose(fp2);

  /*
   * read the data sequence file
   */
  if ( ( fp = fopen ( dbfile , "r" ) ) == NULL ){
    printf( "  %s can not open\n" , dbfile);
    exit(1);
  }
  p=0;
  while (fgets(read_buf,BUFSIZE,fp) != NULL ) {
    sqd[p]->len = seq_read( read_buf, sqd[p]->code, dblength);
    if (lbsd[p]->seq_length != sqd[p]->len)
      printf("Warning: sequence length is different %d %d\n", 
	     lbsd[p]->seq_length, sqd[p]->len);
    p++;
  }
  fclose(fp);

  /*
   * memory allocation (buf_net, dpm and dtw_paa)
   */
  buf_net = (struct NetData *)malloc(sizeof(struct NetData) 
				     * (num_nn + dbsize));
  nn->netd = buf_net;
  vnn->netd = buf_net + num_nn;

  dpm = dpMatch_malloc(dblength, qlength);
}

void nnsearch_m(int query_cnt){

  int i, j, k, p, m;
  float dist, d, initial=INITIAL;
  int   *sp, *ep;

  sp = dpm->start_point;
  ep = dpm->end_point;

  for (i=0 ; i<dbsize ; i++)
    apid[i] = OFF;
  qlbsd->seq_length = qsqd[query_cnt]->len;
  setsegs(segs2, qsqd[query_cnt]->len, calcs);
  for (k=0 ; k<calcs ; k++){
    qlbsd->segd[k].segs = segs2[k];
    getpaa(qsqd[query_cnt]->code, qlbsd, k);
    getelm(warpd->ed[k].elms2, qsqd[query_cnt]->len, segs2[k]);
  }

  if (diff_flg == OFF){
    dpMatch_initialization(dpm, dblength, qsqd[query_cnt]->len);
    for (k=0 ; k<calcs ; k++){
      getelm(warpd->ed[k].elms1, dblength, segs1[k]);
      setsegenv(warpd, sp, ep, segs1[k], segs2[k], dblength, 
		qsqd[query_cnt]->len, k);
    }
  }

  m = num_nn - 1;
  for (i=0 ; i<num_nn ; i++){
    vnn->netd[i].distance = initial;
    nn->netd[i].distance = initial;
  }
  for (i=0 ; i<dbsize ; i++){
    if (diff_flg == ON){
      dpMatch_initialization(dpm, sqd[i]->len, qsqd[query_cnt]->len);
      getelm(warpd->ed[0].elms1, lbsd[i]->seq_length, lbsd[i]->segd[0].segs);
      setsegenv(warpd, sp, ep, lbsd[i]->segd[0].segs, qlbsd->segd[0].segs, 
		lbsd[i]->seq_length, qlbsd->seq_length, 0);
    }
    d = lbs(lbsd[i], qlbsd, warpd, work2, 0, initial);
    apdist[i] = d;
    if (d < vnn->netd[m].distance){
      vnn->netd[m].distance = d;
      vnn->netd[m].id = i;
      qsort(vnn->netd, num_nn, sizeof(struct NetData), change);
    }
  }
  vnn->num = num_nn;

  for (i=0 ; i<num_nn ; i++){
    p = vnn->netd[i].id;
    apid[p] = ON;
    nn->netd[i].id = p;
    if (diff_flg == ON)
      dpMatch_initialization(dpm, sqd[p]->len, qsqd[query_cnt]->len);
    nn->netd[i].distance = dpMatch_constraint(dpm, sqd[p]->code, 
					      qsqd[query_cnt]->code);
#if ACCESS
      n_oid++;
#endif
  }
  nn->num = num_nn;
  qsort(nn->netd, nn->num, sizeof(struct NetData), change);

  for (i=0 ; i<dbsize ; i++){
    if (apid[i] == ON)
      continue;
    if (apdist[i] >= nn->netd[m].distance)
      continue;
    if (diff_flg == ON){
      dpMatch_initialization(dpm, sqd[i]->len, qsqd[query_cnt]->len);
      for (k=1 ; k<calcs ; k++){
	getelm(warpd->ed[k].elms1, lbsd[i]->seq_length, lbsd[i]->segd[k].segs);
        setsegenv(warpd, sp, ep, lbsd[i]->segd[k].segs, qlbsd->segd[k].segs, 
		  lbsd[i]->seq_length, qlbsd->seq_length, k);
      }
    }
    for (k=1 ; k<calcs ; k++){
#if ACCESS
      t_pid[k]++;
#endif
      d = lbs(lbsd[i], qlbsd, warpd, work2, k, nn->netd[m].distance);
      if (d >= nn->netd[m].distance)
	break;
    }
    if (d < nn->netd[m].distance){
      dist = dpMatch_threshold(dpm, sqd[i]->code, qsqd[query_cnt]->code, 
			       nn->netd[m].distance);
      if (dist < nn->netd[m].distance){
	nn->netd[m].distance = dist;
	nn->netd[m].id = i;
	qsort(nn->netd, num_nn, sizeof(struct NetData), change);
      }
#if ACCESS
      n_oid++;
#endif
    }
  }
}

int change(nd1,nd2)
struct NetData *nd1;
struct NetData *nd2;
{
    if (nd1->distance > nd2->distance) return  1;
    if (nd1->distance < nd2->distance) return -1;
    return  0;
}

#if ACCESS
void accinfo(acc_file)
char *acc_file;
{

  int i, k, p, total_segs1, fv, fv5, fv10;
  float ave, tmp;
  FILE       *fp_acc;

  total_segs1=0;
  for (k=0 ; k<calcs ; k++)
    total_segs1 += segs1[k];
  p = dbsize * total_segs1 * 2 * FLOAT_SIZE;
  i = PAGESIZE;
  fv = (int)ceil((float)p / (float)i);
  fv5 = (int)ceil((float)fv / 5.0);
  fv10 = (int)ceil((float)fv / 10.0);

  if ( ( fp_acc = fopen ( acc_file , "w" ) ) == NULL ){
    printf( "  %s can not open\n" , acc_file);
    exit(1);
  }

  ave = (float)t_oid/(float)qsize;
  fprintf(fp_acc,"Average number of page accesses: %f\n", ave+fv);
  fprintf(fp_acc,"(Speed-up factor5: %f, factor10: %f)\n", ave+fv5, ave+fv10);
  fprintf(fp_acc,"Average number of accessed original sequences: %f\n", ave);
  fprintf(fp_acc,"Average number of accessed approximate sequences:\n");
  for (k=0 ; k<calcs ; k++){
    tmp = (float)t_pid[k] / (float)qsize;
    fprintf(fp_acc,"level=%d: %f\n", k, tmp);
  }

  fclose(fp_acc);
}
#endif

void error(char *toolname){
#if ACCESS
  fprintf(stderr,"Usage : %s  indexfile  dbfile  query_file  length  num_neighbors  acc_file\n",toolname);
#else
  fprintf(stderr,"Usage : %s  indexfile  dbfile  query_file  length  num_neighbors\n",toolname);
#endif
    exit(1);
}
